local _G = getfenv(0);

-- Config Data Variables
TipTac_Config = {};
local cfg;

local TT_DefaultConfig = {
	showUnitTip = true,
	showTarget = true,
	showTargetedBy = true,
	pvpName = true,
	colorNameByClass = false,
	classColoredBorder = false,
	updateFreq = 0.5,
	hookTips = true,

	tipTacColor = { 0.1, 0.1, 0.2, 1.0 },
	tipTacBorderColor = { 0.3, 0.3, 0.4, 1.0 },
	tipColor = { 0.1, 0.1, 0.2, 1.0 },
	tipBorderColor = { 0.3, 0.3, 0.4, 1.0 },

-- Az: WIP
--	font = "Fonts\\FRIZQT__.TTF",
--	fontSize = 12;

	tipScale = 1,
	gttScale = 1,

	preFadeTime = 0.6,
	fadeTime = 0.6,

	mouseOffsetX = 0,
	mouseOffsetY = 0,

	healthBar = true,
	healthBarText = true,
	manaBar = false,
	manaBarText = true,
	powerBar = false,
	powerBarText = false,

	showBuffs = true,
	selfBuffsOnly = false,
	showDebuffs = true,
	selfDebuffsOnly = false,
	auraSize = 18,
	auraMaxRows = 2,
	showAuraCooldown = true,

	point = "BOTTOMRIGHT",
	smartAnchor = false,
	mouseAnchor = false,
	mouseAnchorUnitsOnly = true, -- Az: temp option
};

-- Tips modified by TipTac in appearance and scale (You can add to this list if you want to modify more tips)
local TT_TipsToModify = {
	"GameTooltip",
	"ShoppingTooltip1",
	"ShoppingTooltip2",
	"ItemRefTooltip",
	"WorldMapTooltip",
	"AtlasLootTooltip",
};
local tipBackdrop = { bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 8, edgeSize = 16, insets = { left = 4, right = 4, top = 4, bottom = 4 } };

-- "Constants"
local TT_NotSpecified = "Not specified";
local TT_Classification = {
	["worldboss"] = "%s|r (Boss)",
	["rareelite"] = "+%s|r (Rare)",
	["elite"] = "+%s",
	["rare"] = "%s|r (Rare)",
};

-- Colors
local TT_WHITE = "|cffffffff";
local TT_LIGHTGRAY = "|cffc0c0c0";
local TT_GRAY = "|cff808080";
local TT_PURPLE = "|cffff32ff";
local TT_YELLOW = "|cffffff00";
local TT_LIGHTBLUE = "|cff25c1eb";

local TT_ClassColors = {};
for class, color in pairs(RAID_CLASS_COLORS) do
	TT_ClassColors[class] = ("|cff%.2x%.2x%.2x"):format(color.r*255,color.g*255,color.b*255);
end

local TT_ReactionColors = {};
for index, color in ipairs(UnitReactionColor) do
	tinsert(TT_ReactionColors,("|cff%.2x%.2x%.2x"):format(color.r*255,color.g*255,color.b*255));
end

local TT_DifficultyColors = {
	TT_GRAY,
	"|cff40c040",
	TT_YELLOW,
	"|cffff8040",
	"|cffff2020",
};

-- Data Variables (u = unit, e = extras, player = player)
local u, e, player = {}, {};
local auras = {};

--------------------------------------------------------------------------------------------------------
--                                           Frame Handlers                                           --
--------------------------------------------------------------------------------------------------------
function TipTacAnchor_OnLoad(self)
	-- Mod Info
	self.modName = "TipTac";
	self.modVers = GetAddOnMetadata(self.modName,"Version");
	-- Add Slash Command
	SlashCmdList[self.modName] = TipTac_OnSlash;
	_G["SLASH_"..self.modName.."1"] = "/tiptac";
	_G["SLASH_"..self.modName.."2"] = "/tip";
	-- Events
	self:RegisterEvent("PLAYER_LOGIN");
	self:RegisterEvent("PLAYER_LEVEL_UP");
	self:RegisterEvent("VARIABLES_LOADED");
end

-- Events
function TipTacAnchor_OnEvent(self,event)
	-- Player Level Up
	if (event == "PLAYER_LEVEL_UP") then
		player.level = arg1;
	-- Player Login
	elseif (event == "PLAYER_LOGIN") then
		player = { name = UnitName("player"), level = UnitLevel("player") };
	-- Variables Loaded
	elseif (event == "VARIABLES_LOADED") then
		cfg = TipTac_Config;
		TipTac_CheckSettings();
		TipTac_ApplySettings();
		TipTac_HookTips();
		if (cfg.left and cfg.top) then
			TipTacAnchor:ClearAllPoints();
			TipTacAnchor:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT",cfg.left,cfg.top);
		end
	end
end

--------------------------------------------------------------------------------------------------------
--                                           Slash Handling                                           --
--------------------------------------------------------------------------------------------------------
function TipTac_OnSlash(cmd)
	-- Extract Parameters
	local param1, param2 = cmd:match("^([^%s]+)%s*(.*)$");
	param1 = (param1 and param1:lower() or cmd:lower());
	-- Options
	if (param1 == "") then
		local loaded, reason = LoadAddOn("TipTacOptions");
		if (loaded) then
			if (TipTacOptions:IsVisible()) then
				TipTacOptions:Hide();
			else
				TipTacOptions:Show();
			end
		else
			AzMsg("Could not open TicTac Options: |1"..tostring(reason).."|r. Please make sure the addon is enabled from the character selection screen.");
		end
	-- Show Anchor
	elseif (param1 == "anchor") then
		TipTacAnchor:Show();
	-- DEBUG: GTT
	elseif (param1 == "gtt") then
		if (GameTooltip:NumLines() == 0) then
			AzMsg("GameTooltip has 0 lines.");
		else
			for i = 1, GameTooltip:NumLines() do
				AzMsg("|2"..i.."|r = |1"..(_G["GameTooltipTextLeft"..i]:GetText() or "").."   |2"..(_G["GameTooltipTextRight"..i]:GetText() or ""));
			end
		end
	-- Invalid or No Command
	else
		UpdateAddOnMemoryUsage();
		AzMsg("----- |2"..TipTacAnchor.modName.."|r |1"..TipTacAnchor.modVers.."|r ----- |1"..format("%.2f",GetAddOnMemoryUsage(TipTacAnchor.modName)).." |2kb|r -----");
		AzMsg("The following |2parameters|r are valid for |2/tiptac|r or |2/tip|r:");
		AzMsg(" |2anchor|r = Shows the anchor where the tooltip appears");
	end
end
--------------------------------------------------------------------------------------------------------
--                                Global Chat Message Function (Rev 3)                                --
--------------------------------------------------------------------------------------------------------
if (not AZMSG_REV or AZMSG_REV < 3) then
	AZMSG_REV = 3;
	function AzMsg(text)
		DEFAULT_CHAT_FRAME:AddMessage(tostring(text):gsub("|1","|cffffff80"):gsub("|2","|cffffffff"),128/255,192/255,255/255);
	end
end
--------------------------------------------------------------------------------------------------------
--                                              TipTacTip                                             --
--------------------------------------------------------------------------------------------------------

-- Tip OnLoad
function TipTacTip_OnLoad(self)
	self.fadeTime = -1;
	self.lastUpdate = 0;
	self:SetBackdrop(tipBackdrop);
end

-- Tip OnUpdate
function TipTacTip_OnUpdate(self,elapsed)
	self.lastUpdate = (self.lastUpdate + elapsed);
	-- Reposition Tip if Mouse Anchor is enabled (This OnUpdate only happens for units, so no need to check "cfg.mouseAnchorUnitsOnly")
	if (cfg.mouseAnchor) then
		TipTac_AnchorFrameToMouse(self);
	end
	-- Fade Out
	if (self.fadeTime ~= -1) then
		if (GameTooltip:IsVisible() or self.lastUpdate > self.fadeTime + cfg.preFadeTime) then
			self:Hide();
		elseif (self.lastUpdate > cfg.preFadeTime) then
			self:SetAlpha(1 - (self.lastUpdate - cfg.preFadeTime) / self.fadeTime);
		end
	-- Hide Tip if unit is no more, hide GTT only if it is not showing anything new. This seems like a messy way, but don't know of any better.
	-- The GTT:GetUnit() evaluation in the end, is to avoid hiding the GTT when you move to a unitframe which does not call GTT:SetUnit(), but instead uses a custom tip.
	elseif (not UnitIsUnit(u.token,"mouseover") or not GameTooltip:GetUnit() or GameTooltip:GetUnit() ~= UnitName(u.token)) then
		if (not GameTooltip:IsVisible() or GameTooltipTextLeft1:GetText() == u.gttName) then
			self.lastUpdate = 0;
			self.fadeTime = cfg.fadeTime;
			GameTooltip:Hide();
		else
			self:Hide();
		end
	-- Update Tip
	-- Az: Updating the GTT makes TipTac update extras like Beast Lore without moving the mouse off and on, more cpu used I guess?
	-- Although this sounds good, MobInfo2 updates the GTT in a very annoying way, so updating the tooltip with GTT:SetUnit() will
	-- not work as it clears the lines MobInfo2 added, and it doesn't add extra lines by hooking GTT.OnTooltipSetUnit() or GTT:SetUnit()...
	elseif (cfg.updateFreq > 0) and (self.lastUpdate > cfg.updateFreq) then
		self.lastUpdate = 0;
		--GameTooltip:SetUnit(u.token);
		TipTac_ShowUnit(u.token);
	end

	-- Sometimes the GTT changes, that causes its Show() to be called and reset Alpha to 1, if that happens, force it back to 0 here
	if (GameTooltip:GetAlpha() ~= 0 and self:IsVisible()) then
		GameTooltip:SetAlpha(0);
	end
end

--------------------------------------------------------------------------------------------------------
--                                              GTT Hooks                                             --
--------------------------------------------------------------------------------------------------------

-- HOOK: GameTooltip_SetDefaultAnchor (For Re-Anchoring)
function GameTooltip_SetDefaultAnchor(tooltip,parent)
	-- Return if no tooltip or parent
	if (not tooltip or not parent) then
		return;
	end
	-- Position Tip (Mouse/Smart/Normal)
	if (cfg.mouseAnchor) and (not cfg.mouseAnchorUnitsOnly or GameTooltip:GetUnit()) then
		tooltip:SetOwner(parent,"ANCHOR_NONE");
		TipTac_AnchorFrameToMouse(tooltip);
	elseif (cfg.smartAnchor) and (parent ~= UIParent) then
		tooltip:SetOwner(parent,"ANCHOR_RIGHT",cfg.mouseOffsetX,cfg.mouseOffsetY);
	else
		tooltip:SetOwner(parent,"ANCHOR_NONE");
		tooltip:SetPoint(cfg.point,TipTacAnchor);
	end
	tooltip.default = 1;
end

-- HOOK: GTT OnTooltipSetUnit
GameTooltip:HookScript("OnTooltipSetUnit",
function(self)
	if (cfg.showUnitTip) then
		TipTac_ShowUnit(select(2,self:GetUnit()),1);
	end
end
);

-- HOOK: GTT OnUpdate
GameTooltip:HookScript("OnUpdate",
function()
	if (this:GetAlpha() > 0) and (cfg.mouseAnchor) and (not cfg.mouseAnchorUnitsOnly or GameTooltip:GetUnit()) and (this:GetAnchorType() ~= "ANCHOR_CURSOR") then
		TipTac_AnchorFrameToMouse(this);
	end
end
);


-- Az: WIP
-- ChangeNewLinesFont
--[[local function ChangeNewLinesFont(self,...)
	if (self.TipTac_MaxLines < self:NumLines()) then
		--cfg.font = "Interface\\Addons\\sct\\fonts\\Tw_Cen_MT_Bold.TTF";
		cfg.font = "Fonts\\FRIZQT__.TTF";
--AzMsg("Loop |1"..(self.TipTac_MaxLines + 1).."|r to |1"..self:NumLines().."|r.");
		for i = self.TipTac_MaxLines + 1, self:NumLines() do
			_G[self:GetName().."TextLeft"..i]:SetFont(cfg.font,(i > 1 and cfg.fontSize or cfg.fontSize + 2),"NORMAL");
			_G[self:GetName().."TextRight"..i]:SetFont(cfg.font,(i > 1 and cfg.fontSize or cfg.fontSize + 2),"NORMAL");
		end
		self.TipTac_MaxLines = self:NumLines();
		-- call show to resize the frame
		self:Show();
	end
end]]

-- New Tip OnShow
local function TipsOnShow(self,...)
	if (self.OnShow_TipTac) then
		self.OnShow_TipTac(...);
	end

	local unit = select(2,self:GetUnit());
	if (unit) and (UnitIsPlayer(unit)) and (cfg.classColoredBorder) then
		local color = RAID_CLASS_COLORS[select(2,UnitClass(unit))];
		self:SetBackdropBorderColor(color.r,color.g,color.b,1);
	else
		self:SetBackdropBorderColor(unpack(cfg.tipBorderColor)); -- Def: (1,1,1,1)
	end
	self:SetBackdropColor(unpack(cfg.tipColor)); -- Def: For most: (0.1,0.1,0.2), for world objects?: (0,0.2,0.35)

	-- Az: WIP
	--ChangeNewLinesFont(self,...);
end

-- Function to loop through tips to modify and hook
function TipTac_HookTips()
	if (cfg.hookTips) then
		for _, tip in ipairs(TT_TipsToModify) do
			tip = _G[tip];
			if (type(tip) == "table") then
				-- Change Backdrop of the tip
				tip:SetBackdrop(tipBackdrop);
				-- HOOK: Tip OnShow
				tip.TipTac_MaxLines = 0;
				tip.OnShow_TipTac = tip:GetScript("OnShow");
				tip:SetScript("OnShow",TipsOnShow);

				-- Az: WIP
				--hooksecurefunc(tip,"Show",ChangeNewLinesFont);
			end
		end
	end
end

-- This is the function that determines the GTT color for the first line for a unit
local Real_GameTooltip_UnitColor = GameTooltip_UnitColor;
function GameTooltip_UnitColor(unit)
	if (cfg.colorNameByClass) and (UnitIsPlayer(unit)) then
		local color = RAID_CLASS_COLORS[select(2,UnitClass(unit))];
		return color.r, color.g, color.b;
	else
		return Real_GameTooltip_UnitColor(unit);
	end
end

--------------------------------------------------------------------------------------------------------
--                                               Anchor                                               --
--------------------------------------------------------------------------------------------------------

local TT_AnchorPos = {
	{ name = "Anchor Point" },
	{ name = "Top", point = "TOP" },
	{ name = "Top Left", point = "TOPLEFT" },
	{ name = "Top Right", point = "TOPRIGHT" },
	{ name = "Bottom", point = "BOTTOM" },
	{ name = "Bottom Left", point = "BOTTOMLEFT" },
	{ name = "Bottom Right", point = "BOTTOMRIGHT" },
	{ name = "Left", point = "LEFT" },
	{ name = "Right", point = "RIGHT" },
	{ name = "Center", point = "CENTER" },
	{ name = "" },
	{ name = "Anchor Type" },
	{ name = "Mouse Anchor", point = "MOUSE" },
	{ name = "Smart Anchor", point = "SMART" },
};

-- Anchor OnMouseUp
function TipTacAnchor_OnMouseUp()
	if (arg1 == "LeftButton") then
		this:StopMovingOrSizing();
		cfg.left = TipTacAnchor:GetLeft();
		cfg.top = TipTacAnchor:GetTop();
	elseif (arg1 == "RightButton") then
		ToggleDropDownMenu(1,nil,TipTacAnchorDropDown,this:GetName(),0,0);
	end
end

-- DropDown OnLoad
function TipTacAnchor_DropDown_OnLoad(self)
	UIDropDownMenu_Initialize(self,TipTacAnchor_DropDown_Initialize,"MENU");
end

-- DropDown OnClick
function TipTacAnchor_DropDown_OnClick()
	if (this.value == "MOUSE") then
		cfg.mouseAnchor = not cfg.mouseAnchor;
		cfg.smartAnchor = false;
	elseif (this.value == "SMART") then
		cfg.mouseAnchor = false;
		cfg.smartAnchor = not cfg.smartAnchor;
	elseif (this.value) then
		cfg.point = this.value;
	end
end

-- DropDown Init
local info = { func = TipTacAnchor_DropDown_OnClick };
function TipTacAnchor_DropDown_Initialize()
	for index, value in ipairs(TT_AnchorPos) do
		info.text = value.name;
		info.value = value.point;
		info.isTitle = (not value.point and 1 or nil);
		info.disabled = nil;
		info.checked = (cfg and value.point == "SMART" and cfg.smartAnchor) or (cfg and value.point == "MOUSE" and cfg.mouseAnchor) or (cfg and cfg.point == TT_AnchorPos[index].point);
		UIDropDownMenu_AddButton(info);
	end
end

--------------------------------------------------------------------------------------------------------
--                                           Tip Information                                          --
--------------------------------------------------------------------------------------------------------

-- Add Unit Details
function TipTac_AddUnitDetails()
	-- Reaction (Can't rely on UnitPlayerControlled() alone, since it always returns nil on units out of range)
	if (UnitIsPlayer(u.token) or UnitPlayerControlled(u.token)) then
		if (UnitCanAttack(u.token,"player")) then
			if (UnitCanAttack("player",u.token)) then
				u.reaction = TT_ReactionColors[2];
			else
				u.reaction = TT_ReactionColors[3];
			end
		elseif (UnitCanAttack("player",u.token)) then
			u.reaction = TT_ReactionColors[4];
		elseif (UnitIsPVP(u.token) and not UnitIsPVPSanctuary(u.token) and not UnitIsPVPSanctuary("player")) then
			u.reaction = TT_ReactionColors[6];
		else
			u.reaction = TT_LIGHTBLUE;
		end
	elseif (UnitIsTapped(u.token) and not UnitIsTappedByPlayer(u.token)) then
		u.reaction = TT_LIGHTGRAY;
	else
		u.reaction = TT_ReactionColors[UnitReaction(u.token,"player") or 5];
	end
	-- Players
	if (UnitIsPlayer(u.token)) then
		-- class
		u.class, u.classFixed = UnitClass(u.token);
		u.class = " "..(TT_ClassColors[u.classFixed] or TT_WHITE)..u.class;
		-- name
		u.name, u.realm = UnitName(u.token);
		u.line1 = (cfg.colorNameByClass and (TT_ClassColors[u.classFixed] or TT_WHITE) or u.reaction)..(cfg.pvpName and UnitPVPName(u.token) or u.name);
		if (u.realm) and (u.realm ~= "") then
			u.line1 = u.line1.." - "..u.realm;
		end
		-- afk or dnd
		if (UnitIsAFK(u.token)) then
			u.line1 = TT_WHITE.."<AFK> "..u.line1;
		elseif (UnitIsDND(u.token)) then
			u.line1 = TT_WHITE.."<DND> "..u.line1;
		end
		-- guild
		u.guild, u.guildRank = GetGuildInfo(u.token);
		if (u.guild) then
			player.guild = GetGuildInfo("player");
			if (u.guild == player.guild) then
				u.guild = TT_PURPLE.."<"..u.guild..">";
			else
				u.guild = u.reaction.."<"..u.guild..">";
			end
			--u.guild = u.guild.." "..TT_LIGHTGRAY..u.guildRank;
		end
		-- race
		u.race = " "..TT_WHITE..(UnitRace(u.token) or "");
	-- NPCs
	else
		-- name
		u.name = UnitName(u.token);
		u.line1 = u.reaction..u.name;
		-- guild
		u.guild = GameTooltipTextLeft2:GetText();
		u.guildRank = nil;
		if (u.guild) then
			if (u.guild:find("^"..LEVEL.." .+")) then
				u.guild = nil;
			else
				u.guild = u.reaction.."<"..u.guild..">";
			end
		end
		-- race
		u.race = "";
		-- class
		u.class = UnitCreatureFamily(u.token) or UnitCreatureType(u.token);
		if (not u.class or u.class == TT_NotSpecified) then
			u.class = "Unknown";
		end
		u.class = " "..TT_WHITE..u.class;
	end
	-- Target
	if (not cfg.showTarget) then
		u.target = nil;
	else
		u.target = UnitName(u.token.."target");
		if (u.target == "Unknown") and (not UnitExists(u.token.."target")) then
			u.target = nil;
		elseif (u.target) then
			if (u.target == player.name) then
				u.target = TT_WHITE.."[YOU]"
			else
				u.targetReaction = (TT_ReactionColors[UnitReaction(u.token.."target","player")] or TT_LIGHTGRAY);
				if (UnitIsPlayer(u.token.."target")) then
					u.target = u.targetReaction.."["..(TT_ClassColors[select(2,UnitClass(u.token.."target"))] or TT_LIGHTGRAY)..u.target..u.targetReaction.."]";
				else
					u.target = u.targetReaction.."["..u.target.."]";
				end
			end
		end
	end
	-- Level + Classification
	u.level = UnitLevel(u.token);
	u.levelText = (u.level == -1 and "??" or u.level);
	u.classification = UnitClassification(u.token);
	if (TT_Classification[u.classification]) then
		u.levelText = TT_Classification[u.classification]:format(u.levelText);
	end
	if (UnitCanAttack(u.token,"player") or UnitCanAttack("player",u.token)) then
		u.levelText = TipTac_DifficultyLevel(u.level ~= -1 and u.level or 500)..u.levelText.."  ";
	else
		u.levelText = TT_LIGHTGRAY..u.levelText.." ";
	end
	-- Add The Stuff
	TipTacTip:AddLine(u.line1,1,1,1);
	if (u.target) then
		TipTacTip:AddLine("  "..u.target);
		u.raidIcon = GetRaidTargetIndex(u.token.."target");
		if (u.raidIcon) then
			TipTacTip:AddTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons");
			SetRaidTargetIconTexture(TipTacTipTexture1,u.raidIcon);
		end
	end
	if (u.guild) then
		TipTacTip:AddLine(u.guild,1,1,1);
	end
	TipTacTip:AddLine(u.levelText..u.race..u.class,1,1,1);
end

-- Add Extra GTT Lines
function TipTac_AddExtraLines()
	for i = 2, GameTooltip:NumLines() do
		e.lineL = _G["GameTooltipTextLeft"..i]:GetText();
		if (e.lineL) and (e.lineL:find("^"..LEVEL.." .+")) then
			for i = i + 1, GameTooltip:NumLines() do
				e.lineL, e.lineR = _G["GameTooltipTextLeft"..i], _G["GameTooltipTextRight"..i];
				if (e.lineL:GetText() ~= PVP_ENABLED) then
					e.r, e.g, e.b = e.lineL:GetTextColor();
					TipTacTip:AddDoubleLine(e.lineL:GetText(),e.lineR:GetText(),e.r,e.g,e.b,e.lineR:GetTextColor());
				end
			end
			return;
		end
	end
end

-- Add "Targeted By" line
function TipTac_AddTargetedBy()
	e.numParty, e.numRaid = GetNumPartyMembers(), GetNumRaidMembers();
	if (e.numParty > 0 or e.numRaid > 0) then
		e.players, e.counter = "", 0;
		for i = 1, (e.numRaid > 0 and e.numRaid or e.numParty) do
			e.unit = (e.numRaid > 0 and "raid"..i or "party"..i);
			if (UnitIsUnit(e.unit.."target",u.token)) and (not UnitIsUnit(e.unit,"player")) then
				if (mod(e.counter + 3,6) == 0) then
					e.players = e.players.."\n";
				end
				e.color = RAID_CLASS_COLORS[select(2,UnitClass(e.unit))];
				e.players = ("%s|cff%.2x%.2x%.2x%s|r, "):format(e.players,e.color.r*255,e.color.g*255,e.color.b*255,UnitName(e.unit));
				e.counter = (e.counter + 1);
			end
		end
		if (e.players ~= "") then
			TipTacTip:AddLine("Targeted By (|cffffffff"..e.counter.."|r): "..e.players:sub(1,-5));
		end
	end
end

--------------------------------------------------------------------------------------------------------
--                                        Health & Power Bars                                         --
--------------------------------------------------------------------------------------------------------

-- Configures the Health and Power Bars
function TipTac_SetupHealthAndPowerBar()
	u.powerType = UnitPowerType(u.token);
	-- Visibility
	if (cfg.healthBar) then
		TipTacTipHealth:Show();
	else
		TipTacTipHealth:Hide();
	end
	if (cfg.powerBar and u.powerType ~= 0) or (cfg.manaBar and u.powerType == 0) then
		u.powerColor = ManaBarColor[u.powerType] or { r = 0.5, g = 0.5, b = 0.5 };
		TipTacTipPower:SetStatusBarColor(u.powerColor.r,u.powerColor.g,u.powerColor.b);
		TipTacTipPower:Show();
	else
		TipTacTipPower:Hide();
	end
	-- Anchor Frames
	TipTacTipPower:ClearAllPoints();
	TipTacTipHealth:ClearAllPoints();
	if (TipTacTipPower:IsShown()) then
		TipTacTipPower:SetPoint("BOTTOMLEFT",8,9);
		TipTacTipPower:SetPoint("BOTTOMRIGHT",-8,9);
		if (TipTacTipHealth:IsShown()) then
			TipTacTipHealth:SetPoint("BOTTOMLEFT",TipTacTipPower,"TOPLEFT",0,4);
			TipTacTipHealth:SetPoint("BOTTOMRIGHT",TipTacTipPower,"TOPRIGHT",0,4);
		end
	elseif (TipTacTipHealth:IsShown()) then
		TipTacTipHealth:SetPoint("BOTTOMLEFT",8,9);
		TipTacTipHealth:SetPoint("BOTTOMRIGHT",-8,9);
	end
end

-- Update Health & Power
function TipTac_UpdateHealthAndPowerBar()
	-- Health
	if (TipTacTipHealth:IsShown()) then
		u.hp, u.hpMax = UnitHealth(u.token), UnitHealthMax(u.token);
		TipTacTipHealth:SetMinMaxValues(0,u.hpMax);
		TipTacTipHealth:SetValue(u.hp);
		if (cfg.healthBarText) then
			if (u.hpMax ~= 100 or UnitIsUnit(u.token,"player") or UnitIsUnit(u.token,"pet") or UnitPlayerOrPetInParty(u.token) or UnitPlayerOrPetInRaid(u.token))  then
				TipTacTipHealthText:SetText(u.hp.." / "..u.hpMax);
			elseif (MobHealth_PPP) then
				u.ppp = MobHealth_PPP(u.name..":"..u.level);
				if (u.ppp) and (u.ppp > 0) then
					u.hp, u.hpMax = floor(u.ppp * u.hp + 0.5), floor(u.ppp * 100 + 0.5);
					TipTacTipHealthText:SetText(u.hp.." / "..u.hpMax);
				else
					TipTacTipHealthText:SetText(u.hp.."%");
				end
			else
				TipTacTipHealthText:SetText(u.hp.."%");
			end
		end
	end
	-- Power
	if (TipTacTipPower:IsShown()) then
		u.power, u.powerMax = UnitMana(u.token), UnitManaMax(u.token);
		TipTacTipPower:SetMinMaxValues(0,u.powerMax);
		TipTacTipPower:SetValue(u.power);
		if (u.powerType == 0 and cfg.manaBarText) or (u.powerType ~= 0 and cfg.powerBarText) then
			TipTacTipPowerText:SetText(u.power.." / "..u.powerMax);
		else
			TipTacTipPowerText:SetText("");
		end
	end
	-- Cast Bar
	--[[if (1) then
		local spell, rank, displayName, icon, startTime, endTime = UnitCastingInfo(u.token);
		if (not spell) then
			spell, rank, displayName, icon, startTime, endTime = UnitChannelInfo(u.token);
			if (not spell) then
				return;
			end
		end
		startTime, endTime = startTime/1000, endTime/1000;
		TipTacTipHealth:SetMinMaxValues(startTime,endTime);
		TipTacTipHealth:SetValue(GetTime());
		TipTacTipHealthText:SetFormattedText("%s %d%%",spell,((GetTime()-startTime) / (endTime-startTime) * 100));
	end]]
end

--------------------------------------------------------------------------------------------------------
--                                       Auras - Buffs & Debuffs                                      --
--------------------------------------------------------------------------------------------------------

function TipTac_CreateAura()
	local aura = CreateFrame("Frame",nil,TipTacTip);
	aura:SetWidth(cfg.auraSize);
	aura:SetHeight(cfg.auraSize);
	aura.count = aura:CreateFontString(nil,"OVERLAY");
	aura.count:SetPoint("BOTTOMRIGHT",1,0);
	aura.count:SetFont("Fonts\\FRIZQT__.TTF",(cfg.auraSize / 2),"OUTLINE");
	aura.icon = aura:CreateTexture(nil,"BACKGROUND");
	aura.icon:SetAllPoints();
	aura.cooldown = CreateFrame("Cooldown",nil,aura,"CooldownFrameTemplate");
	aura.cooldown:SetReverse(1);
	aura.cooldown:SetAllPoints();
	aura.border = aura:CreateTexture(nil,"OVERLAY");
	aura.border:SetPoint("TOPLEFT",-1,1);
	aura.border:SetPoint("BOTTOMRIGHT",1,-1);
	aura.border:SetTexture("Interface\\Buttons\\UI-Debuff-Overlays");
	aura.border:SetTexCoord(0.296875,0.5703125,0,0.515625);
	tinsert(auras,aura);
	return aura;
end

function TipTac_SetupAuras()
	local index = 1;
	local aura;
	local maxAuras = floor(TipTacTip:GetWidth() / (cfg.auraSize + 1));
	local iconTexture, count, debuffType, duration, timeLeft;
	-- Get Buffs
	if (cfg.showBuffs) then
		iconTexture, count, duration, timeLeft = select(3,UnitBuff(u.token,index,cfg.selfBuffsOnly));
		while (iconTexture) do
			-- Get Aura Frame
			aura = auras[index] or TipTac_CreateAura();
			-- Anchor It
			aura:ClearAllPoints();
			if (index == 1) then
				aura:SetPoint("BOTTOMLEFT",TipTacTip,"TOPLEFT");
			elseif (maxAuras == 1) or (mod(index,maxAuras) == 1) then
				aura:SetPoint("BOTTOM",auras[index - maxAuras],"TOP",0,1);
			else
				aura:SetPoint("LEFT",auras[index - 1],"RIGHT",1,0);
			end
			-- Cooldown
			if (cfg.showAuraCooldown) and (duration and duration > 0 and timeLeft and timeLeft > 0) then
				aura.cooldown:SetCooldown(GetTime() + timeLeft - duration,duration);
			else
				aura.cooldown:Hide();
			end
			-- Set Texture + Count
			aura.icon:SetTexture(iconTexture);
			aura.count:SetText(count and count > 1 and count or "");
			-- Border
			aura.border:Hide();
			-- Show + Next, Break if exceed max desired rows of aura
			aura:Show();
			index = (index + 1);
			if (index / maxAuras > cfg.auraMaxRows) then
				break;
			end
			iconTexture, count, duration, timeLeft = select(3,UnitBuff(u.token,index,cfg.selfBuffsOnly));
		end
	end
	-- Get Debuffs
	if (cfg.showDebuffs) and (index / maxAuras <= cfg.auraMaxRows) then
		local firstIndex;
		local buffCount, debuffIndex = (index - 1), 1;
		local lineOffset = floor(buffCount / maxAuras);
		local buffsOnFirstLine = (maxAuras - mod(buffCount,maxAuras));
		iconTexture, count, debuffType, duration, timeLeft = select(3,UnitDebuff(u.token,debuffIndex,cfg.selfDebuffsOnly));
		while (iconTexture) do
			-- Get Aura Frame
			aura = auras[index] or TipTac_CreateAura();
			-- Anchor It
			aura:ClearAllPoints();
			if (debuffIndex <= buffsOnFirstLine) and (debuffIndex == 1) then
				aura:SetPoint("BOTTOMRIGHT",TipTacTip,"TOPRIGHT",0,(cfg.auraSize + 1) * lineOffset);
			elseif (mod(index - 1,maxAuras) == 0) then
				aura:SetPoint("BOTTOMRIGHT",TipTacTip,"TOPRIGHT",0,(cfg.auraSize + 1) * floor(index / maxAuras));
			else
				aura:SetPoint("RIGHT",auras[index - 1],"LEFT",-1,0);
			end
			-- Cooldown
			if (cfg.showAuraCooldown) and (duration and duration > 0 and timeLeft and timeLeft > 0) then
				aura.cooldown:SetCooldown(GetTime() + timeLeft - duration,duration);
			else
				aura.cooldown:Hide();
			end
			-- Set Texture + Count
			aura.icon:SetTexture(iconTexture);
			aura.count:SetText(count and count > 1 and count or "");
			-- Border
			local color = DebuffTypeColor[debuffType or "none"];
			aura.border:SetVertexColor(color.r,color.g,color.b);
			aura.border:Show();
			-- Show + Next, Break if exceed max desired rows of aura
			aura:Show();
			index = (index + 1);
			debuffIndex = (debuffIndex + 1);
			if (index / maxAuras > cfg.auraMaxRows) then
				break;
			end
			iconTexture, count, debuffType, duration, timeLeft = select(3,UnitDebuff(u.token,debuffIndex,cfg.selfDebuffsOnly));
		end
	end
	-- Hide the rest
	for i = index, #auras do
		auras[i]:Hide();
	end
end

--------------------------------------------------------------------------------------------------------
--                                                Misc                                                --
--------------------------------------------------------------------------------------------------------

-- Defaults Void Settings
function TipTac_CheckSettings()
	for option, value in pairs(TT_DefaultConfig) do
		if (cfg[option] == nil) then
			cfg[option] = value;
		end
	end
end

-- Apply Settings
function TipTac_ApplySettings()
	-- Set Scale
	TipTacTip:SetScale(cfg.tipScale);
	for _, tip in ipairs(TT_TipsToModify) do
		tip = _G[tip];
		if (type(tip) == "table") then
			tip:SetScale(cfg.gttScale);
			tip.TipTac_MaxLines = 0;
		end
	end
	-- If disabled, hide auras, else set their size
	for _, aura in ipairs(auras) do
		if (cfg.showBuffs or cfg.showDebuffs) then
			aura:SetWidth(cfg.auraSize);
			aura:SetHeight(cfg.auraSize);
			aura.count:SetFont("Fonts\\FRIZQT__.TTF",(cfg.auraSize / 2),"OUTLINE");
		else
			aura:Hide();
		end
	end
	-- Set TipTacTip Appearance
	TipTacTip:SetBackdropColor(unpack(cfg.tipTacColor));
	if (not cfg.classColoredBorder) then
		TipTacTip:SetBackdropBorderColor(unpack(cfg.tipTacBorderColor));
	end
	-- Health Bar Text
	if (cfg.healthBarText) then
		TipTacTipHealthText:Show();
	else
		TipTacTipHealthText:Hide();
	end
end

-- Returns the correct difficulty color compared to the player
function TipTac_DifficultyLevel(level)
	level = (level - player.level);
	if (level > 4) then
		return TT_DifficultyColors[5]; -- red
	elseif (level > 2) then
		return TT_DifficultyColors[4]; -- orange
	elseif (level >= -2) then
		return TT_DifficultyColors[3]; -- yellow
	elseif (level >= -GetQuestGreenRange()) then
		return TT_DifficultyColors[2]; -- green
	else
		return TT_DifficultyColors[1]; -- gray
	end
end

-- Anchor any given frame to mouse position
function TipTac_AnchorFrameToMouse(frame)
	local x, y = GetCursorPosition();
	local effScale = frame:GetEffectiveScale();
	frame:ClearAllPoints();
	frame:SetPoint(cfg.point,UIParent,"BOTTOMLEFT",(x / effScale + cfg.mouseOffsetX),(y / effScale + cfg.mouseOffsetY));
end

-- Show Unit
function TipTac_ShowUnit(unitToken,first)
	-- Init
	if (first) then
		u.token = unitToken;
		u.gttName = GameTooltipTextLeft1:GetText();
		GameTooltip_SetDefaultAnchor(TipTacTip,GameTooltip:GetOwner());
	else
		TipTacTip:ClearLines();
	end
	-- Construct the TipTacTip
	TipTac_AddUnitDetails();
	TipTac_AddExtraLines();
	if (cfg.showTargetedBy) then
		TipTac_AddTargetedBy();
	end
	if (cfg.showBuffs or cfg.showDebuffs) then
		TipTac_SetupAuras();
	end
	if (first) or (UnitPowerType(u.token) ~= u.powerType) then
		TipTac_SetupHealthAndPowerBar();
	end
	TipTac_UpdateHealthAndPowerBar();
	-- Border
	if (first) and (cfg.classColoredBorder) then
		if (UnitIsPlayer(unitToken)) then
			local color = RAID_CLASS_COLORS[u.classFixed];
			TipTacTip:SetBackdropBorderColor(color.r,color.g,color.b,1);
		else
			TipTacTip:SetBackdropBorderColor(unpack(cfg.tipBorderColor));
		end
	end
	-- Hide GTT & Reset Fade
	if (first) then
		GameTooltip:SetAlpha(0);
		TipTacTip.fadeTime = -1;
	end
	-- Show & Adjust Height
	TipTacTip:Show();
	TipTacTip:SetHeight(TipTacTip:GetHeight() + (TipTacTipHealth:IsShown() and 10 or 0) + (TipTacTipPower:IsShown() and 10 or 0));
end