local _G = getfenv(0);

local cfg = TipTac_Config;

local TTO_AnchorPos = {
	{ name = "Top", value = "TOP" },
	{ name = "Top Left", value = "TOPLEFT" },
	{ name = "Top Right", value = "TOPRIGHT" },
	{ name = "Bottom", value = "BOTTOM" },
	{ name = "Bottom Left", value = "BOTTOMLEFT" },
	{ name = "Bottom Right", value = "BOTTOMRIGHT" },
	{ name = "Left", value = "LEFT" },
	{ name = "Right", value = "RIGHT" },
	{ name = "Center", value = "CENTER" },
};

-- Options
local TTO_Frames = {};
local TTO_Options = {
	-- General
	{
		[0] = "General",
		{ type = "Check", var = "showUnitTip", label = "Show TipTac Unit Tips", tip = "If you disable this option, you will only really take advantage of the anchor and scale options in TipTac" },
		{ type = "Check", var = "pvpName", label = "Use PvP Title If Available", tip = "Show PvP title for players if available" },
		{ type = "Check", var = "showTarget", label = "Show Who the Unit Targets", tip = "This will let you see who the unit targets" },
		{ type = "Check", var = "showTargetedBy", label = "Show Who Targets the Unit", tip = "When in a raid or party, it will show who from your group is targeting the unit" },
		{ type = "Check", var = "colorNameByClass", label = "Color Player Names by Class Color", tip = "With this option on, player names are colored by their class color, otherwise they will be colored by reaction", y = 12 },
		{ type = "Check", var = "classColoredBorder", label = "Color Tip Border by Class Color", tip = "For players, with this option on, the border color will be colored to match the class color of the unit" },
		{ type = "Slider", var = "updateFreq", label = "Tip Update Frequency", min = 0, max = 5, step = 0.05, y = 8 },
		{ type = "Check", var = "hookTips", label = "Hook Special Tips", tip = "Hook special tips to modify in appearance. Changing this options requires a UI reload to take effect.", y = 12 },
 	},
	-- Colors
	{
		[0] = "Colors",
		{ type = "Color", var = "tipTacColor", label = "TipTac Background Color" },
		{ type = "Color", var = "tipTacBorderColor", label = "TipTac Border Color" },
		{ type = "Color", var = "tipColor", label = "Tip Background Color", y = 12 },
		{ type = "Color", var = "tipBorderColor", label = "Tip Border Color" },
	},
	-- Az: WIP
	--Font
--[[	{
		[0] = "Font",
		{ type = "Slider", var = "fontSize", label = "Tip Font Size", min = 6, max = 29, step = 1 },
	},]]
	-- Scaling
	{
		[0] = "Scaling",
		{ type = "Slider", var = "tipScale", label = "TipTacTip Scale", min = 0.2, max = 3, step = 0.05 },
		{ type = "Slider", var = "gttScale", label = "GameTooltip Scale", min = 0.2, max = 3, step = 0.05 },
	},
	-- Fading
	{
		[0] = "Fading",
		{ type = "Slider", var = "preFadeTime", label = "Pre Fade Time", min = 0, max = 5, step = 0.05 },
		{ type = "Slider", var = "fadeTime", label = "Fade Out Time", min = 0, max = 5, step = 0.05 },
	},
	-- Bars
	{
		[0] = "Bars",
		{ type = "Check", var = "healthBar", label = "Show Health Bar", tip = "Will show a health bar of the unit. This supports MobHealth, and will show estimated health if installed and available for that unit." },
		{ type = "Check", var = "healthBarText", label = "Show Health Bar Text", tip = "Show text on health bar" },
		{ type = "Check", var = "manaBar", label = "Show Mana Bar", tip = "If the unit has mana, a mana bar will be shown.", y = 14 },
		{ type = "Check", var = "manaBarText", label = "Show Mana Bar Text", tip = "Show text on mana bar" },
		{ type = "Check", var = "powerBar", label = "Show Energy, Rage or Focus Bar", tip = "If the unit has either an energy, rage or focus bar, it will be shown.", y = 14 },
		{ type = "Check", var = "powerBarText", label = "Show Energy, Rage or Focus Bar Text", tip = "Show text on power bar" },
	},
	-- Auras
	{
		[0] = "Auras",
		{ type = "Check", var = "showBuffs", label = "Show Unit Buffs", tip = "Show buffs of the unit" },
		{ type = "Check", var = "selfBuffsOnly", label = "Show Castable Buffs Only", tip = "This will filter out and not display any buffs you cannot cast" },
		{ type = "Check", var = "showDebuffs", label = "Show Unit Debuffs", tip = "Show debuffs of the unit", y = 12 },
		{ type = "Check", var = "selfDebuffsOnly", label = "Show Removable Debuffs Only", tip = "This will filter out and not display any debuffs you cannot remove" },
		{ type = "Slider", var = "auraSize", label = "Aura Icon Dimension", min = 8, max = 40, step = 1, y = 12 },
		{ type = "Slider", var = "auraMaxRows", label = "Max Aura Rows", min = 1, max = 8, step = 1 },
		{ type = "Check", var = "showAuraCooldown", label = "Show Cooldown Models", tip = "With this option on, you will see a visual progress of the time left on the buff. Only works for buffs you cast.", y = 8 },
	},
	-- Anchor
	{
		[0] = "Anchor",
		{ type = "Check", var = "mouseAnchorUnitsOnly", label = "Only mouse anchor unit tips", tip = "When enabled, non unit tips will still be anchored to the normal anchor" },
--[[		{ type = "DropDown", var = "unitAnchorType", label = "Unit Anchor Type", list = { "Normal", "Mouse", "Smart" } },
		{ type = "DropDown", var = "unitPoint", label = "Unit Anchor Point", list = TTO_AnchorPos },

		{ type = "DropDown", var = "nonUnitAnchorType", label = "Non-Unit Anchor Type", list = { "Normal", "Mouse", "Smart" }, y = 12 },
		{ type = "DropDown", var = "nonUnitPoint", label = "Non-Unit Anchor Point", list = TTO_AnchorPos },]]
	},
	-- Offsets
	{
		[0] = "Offsets",
		{ type = "Slider", var = "mouseOffsetX", label = "Mouse Anchor X Offset", min = -200, max = 200, step = 1 },
		{ type = "Slider", var = "mouseOffsetY", label = "Mouse Anchor Y Offset", min = -200, max = 200, step = 1 },
	},
};

local activePage = 1;
local updatingOptions;

--------------------------------------------------------------------------------------------------------
--                                           Frame Scripts                                            --
--------------------------------------------------------------------------------------------------------
function TipTacOptions_OnLoad(self)
	tinsert(UISpecialFrames,self:GetName());
	TipTacOptionsVers:SetText("|cffffff80"..TipTacAnchor.modVers);
	if (TipTacAnchor) then
		TipTacOptionsAnchor:Show();
	end
	if (cfg.optionsLeft) and (cfg.optionsBottom) then
		self:SetPoint("BOTTOMLEFT",UIParent,"BOTTOMLEFT",cfg.optionsLeft,cfg.optionsBottom);
	end
end

-- OnShow
function TipTacOptions_OnShow(self)
	TipTacOptions_BuildPage();
	TipTacOptions_List_Update();
end

-- MouseUp
function TipTacOptions_OnMouseUp(self,button)
	self:StopMovingOrSizing();
	cfg.optionsLeft = self:GetLeft();
	cfg.optionsBottom = self:GetBottom();
end

--------------------------------------------------------------------------------------------------------
--                                            Slider Frame                                            --
--------------------------------------------------------------------------------------------------------

-- Slider Value Changed
local function Slider_OnValueChanged(self)
	if (not updatingOptions) then
		-- Tooltip
		if (GameTooltip:IsOwned(self)) then
			GameTooltip:SetText(format("%.2f",self:GetValue()));
		end
		-- Change Edit, Save & Apply
		self:GetParent().edit:SetNumber(self:GetValue());
		cfg[TTO_Options[activePage][self:GetParent().id].var] = self:GetValue();
		TipTac_ApplySettings();
	end
end

-- OnMouseWheel
local function Slider_OnMouseWheel(self)
	self:SetValue(self:GetValue() + TTO_Options[activePage][self:GetParent().id].step * arg1);
end

-- EditBox Enter Pressed
local function SliderEdit_OnEnterPressed(self)
	self:GetParent().slider:SetValue(self:GetNumber());
end

-- New Slider
local function Slider_New(index)
	local f = CreateFrame("Frame",nil,TipTacOptions);
	f:SetWidth(292);
	f:SetHeight(30);

	f.edit = CreateFrame("EditBox","TipTacOptionsEdit"..index,f,"InputBoxTemplate");
	f.edit:SetWidth(45);
	f.edit:SetHeight(21);
	f.edit:SetPoint("LEFT");
	f.edit:SetScript("OnEnterPressed",function(self) self:GetParent().slider:SetValue(self:GetNumber()); end);
	f.edit:SetAutoFocus(false);
	f.edit:SetMaxLetters(4);
	f.edit:SetFrameLevel(f.edit:GetFrameLevel() + 5);

	f.slider = CreateFrame("Slider","TipTacOptionsSlider"..index,f,"OptionsSliderTemplate");
	f.slider:SetPoint("TOPLEFT",f.edit,"TOPRIGHT",5,0);
	f.slider:SetPoint("BOTTOMRIGHT",0,5);
	f.slider:SetScript("OnValueChanged",Slider_OnValueChanged);
	f.slider:SetScript("OnMouseWheel",Slider_OnMouseWheel);
	f.slider:SetScript("OnEnter",function(self) GameTooltip:SetOwner(self,"ANCHOR_TOPRIGHT"); GameTooltip:SetText(format("%.2f",self:GetValue())); end);
	f.slider:SetScript("OnLeave",function(self) GameTooltip:Hide(); end);
	f.slider:EnableMouseWheel(1);

	f.text = _G["TipTacOptionsSlider"..index.."Text"];
	f.low = _G["TipTacOptionsSlider"..index.."Low"];
	f.high = _G["TipTacOptionsSlider"..index.."High"];

	return f;
end

--------------------------------------------------------------------------------------------------------
--                                           Check Buttons                                            --
--------------------------------------------------------------------------------------------------------

local function CheckButton_OnEnter(self)
	GameTooltip:SetOwner(self,"ANCHOR_RIGHT");
	GameTooltip:SetText(TTO_Options[activePage][self.id].tip,nil,nil,nil,nil,1);
end

local function CheckButton_OnClick(self)
	cfg[TTO_Options[activePage][self.id].var] = (self:GetChecked() == 1);
	TipTac_ApplySettings();
end

-- New CheckButton
local function CheckButton_New(index)
	local f = CreateFrame("CheckButton","TipTacCheckButton"..index,TipTacOptions,"OptionsCheckButtonTemplate");
	f:SetWidth(26);
	f:SetHeight(26);
	f:SetScript("OnEnter",CheckButton_OnEnter);
	f:SetScript("OnClick",CheckButton_OnClick);

	f.text = _G["TipTacCheckButton"..index.."Text"];

	return f;
end

--------------------------------------------------------------------------------------------------------
--                                           Color Buttons                                            --
--------------------------------------------------------------------------------------------------------

-- Color Picker Function
local function ColorButton_ColorPickerFunc(prevVal,...)
	local r, g, b, a;
	if (prevVal) then
		r, g, b, a  = unpack(prevVal);
	else
		r, g, b = ColorPickerFrame:GetColorRGB();
		a = (1 - OpacitySliderFrame:GetValue());
	end
	ColorPickerFrame.frame.texture:SetVertexColor(r,g,b,a);
	cfg[TTO_Options[activePage][ColorPickerFrame.frame.id].var] = { r, g, b, a };
	TipTac_ApplySettings();
end

-- OnClick
local function ColorButton_OnClick(self,button)
	local r, g, b, a = unpack(cfg[TTO_Options[activePage][self.id].var]);

	ColorPickerFrame.frame = self;
	ColorPickerFrame.func = ColorButton_ColorPickerFunc;
	ColorPickerFrame.cancelFunc = ColorButton_ColorPickerFunc;
	ColorPickerFrame.opacityFunc = ColorButton_ColorPickerFunc;
	ColorPickerFrame.hasOpacity = true;
	ColorPickerFrame.opacity = (1 - a);
	ColorPickerFrame.previousValues = { r, g, b, a };

	OpacitySliderFrame:SetValue(1 - a);
	ColorPickerFrame:SetColorRGB(r,g,b);
	ColorPickerFrame:Show();
end

-- OnEnter
local function ColorButton_OnEnter(self)
	self.border:SetVertexColor(1,1,0);
	if (TTO_Options[activePage][self.id].tip) then
		GameTooltip:SetOwner(self,"ANCHOR_RIGHT");
		GameTooltip:SetText(TTO_Options[activePage][self.id].tip,nil,nil,nil,nil,1);
	end
end

-- New ColorButton
local function ColorButton_New()
	local f = CreateFrame("Button",nil,TipTacOptions);
	f:SetWidth(18);
	f:SetHeight(18);
	f:SetScript("OnEnter",ColorButton_OnEnter);
	f:SetScript("OnLeave",function(self) self.border:SetVertexColor(1,1,1); GameTooltip:Hide() end)
	f:SetScript("OnClick",ColorButton_OnClick);

	f.texture = f:CreateTexture();
	f.texture:SetPoint("TOPLEFT",-1,1);
	f.texture:SetPoint("BOTTOMRIGHT",1,-1);
	f.texture:SetTexture("Interface\\ChatFrame\\ChatFrameColorSwatch");
	f:SetNormalTexture(f.texture);

	f.border = f:CreateTexture(nil,"BACKGROUND");
	f.border:SetPoint("TOPLEFT");
	f.border:SetPoint("BOTTOMRIGHT");
	f.border:SetTexture(1,1,1,1);

	f.text = f:CreateFontString(nil,"ARTWORK","GameFontNormal");
	f.text:SetPoint("LEFT",f,"RIGHT",4,1);

	return f;
end

--------------------------------------------------------------------------------------------------------
--                                         Initialize Options                                         --
--------------------------------------------------------------------------------------------------------

-- Get Frame
function TipTacOptions_GetFrame(type,id,index)
	local frame;
	-- Find existsing frame
	if (TTO_Frames[type]) and (TTO_Frames[type][index]) then
		frame = TTO_Frames[type][index];
	-- Create new frame
	else
		if (type == "Slider") then
			frame = Slider_New(index);
		elseif (type == "Check") then
			frame = CheckButton_New(index);
		elseif (type == "Color") then
			frame = ColorButton_New(index);
		end
		-- Add Frame
		if (not TTO_Frames[type]) then
			TTO_Frames[type] = {};
		end
		tinsert(TTO_Frames[type],frame);
	end
	-- Return Frame
	frame.id = id;
	return frame;
end

local xOffsets = { ["Check"] = 18, ["Slider"] = 28, ["Color"] = 24 };
local yExtraOffsets = { ["Check"] = 0, ["Slider"] = 4, ["Color"] = 8 };

-- Build Page
function TipTacOptions_BuildPage()
	local frame;
	local frameUseCount = {};
	local yOffset = -50;
	-- Loop Through Options
	updatingOptions = 1;
	for index, table in ipairs(TTO_Options[activePage]) do
		-- Init the Frame
		frameUseCount[table.type] = (frameUseCount[table.type] or 0) + 1;
		frame = TipTacOptions_GetFrame(table.type,index,frameUseCount[table.type]);
		-- Anchor the Frame
		frame:ClearAllPoints();
		frame:SetPoint("TOPLEFT",xOffsets[table.type] + (table.x or 0),yOffset - (table.y or 0));
		yOffset = (yOffset - frame:GetHeight() - yExtraOffsets[table.type] - (table.y or 0));
		-- Setup the Frame
		if (table.type == "Slider") then
			frame.slider:SetMinMaxValues(table.min,table.max);
			frame.slider:SetValueStep(table.step);
			frame.slider:SetValue(cfg[table.var]);
			frame.edit:SetNumber(cfg[table.var]);
			frame.text:SetText(table.label);
			frame.low:SetText(table.min);
			frame.high:SetText(table.max);
		elseif (table.type == "Check") then
			frame.text:SetText(table.label);
			frame:SetHitRectInsets(0,frame.text:GetWidth() * -1,0,0);
			frame:SetChecked(cfg[table.var]);
		elseif (table.type == "Color") then
			frame.text:SetText(table.label);
			frame:SetHitRectInsets(0,frame.text:GetWidth() * -1,0,0);
			frame.texture:SetVertexColor(unpack(cfg[table.var]));
		end
		-- Show
		frame:Show();
	end
	-- Hide Unused Frames
	for type, table in pairs(TTO_Frames) do
		for i = (frameUseCount[type] or 0) + 1, #table do
			table[i]:Hide();
		end
	end
	-- Reset
	updatingOptions = nil;
end

--------------------------------------------------------------------------------------------------------
--                                             Page List                                              --
--------------------------------------------------------------------------------------------------------

function TipTacOptions_List_Update()
	TipTacOptionsList:SetHeight(18 * #TTO_Options + 36);
	local button;
	for index, table in ipairs(TTO_Options) do
		button = _G["TipTacOptionsListEntry"..index] or CreateFrame("Button","TipTacOptionsListEntry"..index,TipTacOptionsList,"TipTacOptionsListEntryTemplate");
		_G["TipTacOptionsListEntry"..index.."Text"]:SetText((index == activePage and "|cffffffff"  or "")..table[0]);
		button.page = index;
		button:ClearAllPoints();
		if (index == 1) then
			button:SetPoint("TOPLEFT",TipTacOptionsListOutline,"TOPLEFT",5,-6);
		else
			button:SetPoint("TOPLEFT","TipTacOptionsListEntry"..(index - 1),"BOTTOMLEFT");
		end
	end
end

function TipTacOptions_List_OnClick(self,button)
	activePage = self.page;
	PlaySound("igMainMenuOptionCheckBoxOn");
	TipTacOptions_BuildPage();
	TipTacOptions_List_Update();
end

--------------------------------------------------------------------------------------------------------
--                                                Misc                                                --
--------------------------------------------------------------------------------------------------------
function TipTacOptions_Anchor()
	if (TipTacAnchor:IsVisible()) then
		TipTacAnchor:Hide();
	else
		TipTacAnchor:Show();
	end
end

function TipTacOptions_Reset()
	for index, table in ipairs(TTO_Options[activePage]) do
		cfg[table.var] = nil;
	end
	TipTac_CheckSettings();
	TipTac_ApplySettings();
	TipTacOptions_BuildPage();
end